﻿using System.Text;
using Microsoft.Extensions.Logging;

namespace GlaxLogger;

public class Logger : ILogger, IDisposable, IAsyncDisposable
{
    private LogLevel _filterLevel;
    private readonly FileStream _allMessageLogfile, _filteredLogfile;
    private readonly ConsoleColor _defaultForegroundColor = Console.ForegroundColor;
    private readonly ConsoleColor _defaultBackgroundColor = Console.BackgroundColor;
    private readonly TextWriter _consoleOut;

    public Logger(LogLevel filteredLevel = LogLevel.Warning, string? outputFolderPath = null, TextWriter? consoleOut = null)
    {
        this._filterLevel = filteredLevel;
        string logFolderPath = outputFolderPath ?? Environment.CurrentDirectory;
        this._filteredLogfile = new FileStream(Path.Join(logFolderPath, $"{DateTime.Now:yyyy-MM-dd HH.mm.ss}-filtered.log"), FileMode.Create); 
        this._allMessageLogfile = new FileStream(Path.Join(logFolderPath, $"{DateTime.Now:yyyy-MM-dd HH.mm.ss}.log"), FileMode.Create);
        this._consoleOut = consoleOut ?? Console.Out;
    }
    
    public void Log<TState>(LogLevel logLevel, EventId eventId, TState state, Exception? exception, Func<TState, Exception?, string> formatter)
    {
        string logLevelStr = logLevel.ToString()[..3].ToUpper();
        string timeStr = $"[{DateTime.UtcNow:HH:mm:ss.fff}]";
        string messageStr = formatter.Invoke(state, exception);
        string fullMessage = $"{logLevelStr} {timeStr} {messageStr.Replace("\n","\n\t")}\n\r";
        byte[] fullMessageBytes = Encoding.UTF8.GetBytes(fullMessage);
        _allMessageLogfile.Write(fullMessageBytes);
        if (!IsEnabled(logLevel))
            return;
        _filteredLogfile.Write(fullMessageBytes);
        Console.ForegroundColor = ForegroundColorForLogLevel(logLevel);
        Console.BackgroundColor = BackgroundColorForLogLevel(logLevel);
        _consoleOut.Write(logLevelStr);
        Console.ResetColor();
        // ReSharper disable once LocalizableElement
        _consoleOut.Write($" {timeStr} ");
        _consoleOut.WriteLine(messageStr);
    }

    public bool IsEnabled(LogLevel logLevel)
    {
        return logLevel >= _filterLevel;
    }

    public void UpdateLogLevel(LogLevel filterLevel)
    {
        this._filterLevel = filterLevel;
        Log(LogLevel.None, new EventId(), $"Change LogLevel to {filterLevel}", null, (s, exception) => s);
    }

    public IDisposable? BeginScope<TState>(TState state) where TState : notnull
    {
        return null;
    }
    
    private ConsoleColor ForegroundColorForLogLevel(LogLevel logLevel)
    {
        return logLevel switch
        {
            LogLevel.Error or LogLevel.Critical => ConsoleColor.Black,
            LogLevel.Debug => ConsoleColor.Black,
            LogLevel.Information => ConsoleColor.White,
            _ => _defaultForegroundColor
        };
    }
    
    private ConsoleColor BackgroundColorForLogLevel(LogLevel logLevel)
    {
        return logLevel switch
        {
            LogLevel.Error or LogLevel.Critical => ConsoleColor.Red,
            LogLevel.Debug => ConsoleColor.Yellow,
            LogLevel.Information => ConsoleColor.Black,
            _ => _defaultBackgroundColor
        };
    }

    public void Dispose()
    {
        _allMessageLogfile.Dispose();
        _filteredLogfile.Dispose();
        _consoleOut.Dispose();
    }

    public async ValueTask DisposeAsync()
    {
        await _allMessageLogfile.DisposeAsync();
        await _filteredLogfile.DisposeAsync();
        await _consoleOut.DisposeAsync();
    }
}